import { z } from 'zod';
export * from './errors.js';
export * from './pg.js';
/**
 * @description
 * Takes an Array<V>, and a grouping function,
 * and returns a Map of the array grouped by the grouping function.
 *
 * @param list An array of type V.
 * @param keyGetter A Function that takes the the Array type V as an input, and returns a value of type K.
 *                  K is generally intended to be a property key of V.
 *
 * @returns Map of the array grouped by the grouping function.
 */
export function groupBy(list, keyGetter) {
    const map = new Map();
    list.forEach((item) => {
        const key = keyGetter(item);
        const collection = map.get(key);
        if (!collection) {
            map.set(key, [item]);
        }
        else {
            collection.push(item);
        }
    });
    return map;
}
export const interleave = (arr, x) => arr.flatMap((e) => [e, x]).slice(0, -1);
/**
 * https://stackoverflow.com/a/56592365/21643056
 */
export function pick(obj, paths) {
    const ret = Object.create(null);
    for (const k of paths) {
        ret[k] = obj[k];
    }
    return ret;
}
export const parseWithDefault = (value, defaultValue) => {
    if (!value)
        return defaultValue;
    try {
        return JSON.parse(value);
    }
    catch (error) {
        return defaultValue;
    }
};
export function implement() {
    return {
        with: (schema) => z.object(schema),
    };
}
/**
 * isIn
 * Used in e.g. schema refinement to check if a client value is in a list of valid values.
 */
export const isIn = (val, array) => {
    if (typeof val === 'string')
        return array.includes(val);
    return val.every((v) => array.includes(v));
};
/**
 * undefinedOrLte
 * Used in e.g. schema refinement to check if a client value is less than or equal to a max value.
 */
export const undefinedOrLte = (val, max) => {
    if (!val)
        return true;
    return val <= max;
};
